package com.zy.bms.config.mqtt;

import com.alibaba.fastjson.JSON;
import com.zy.bms.common.Constant;
import com.zy.bms.common.exception.ApiRuntimeException;
import com.zy.bms.model.*;
import com.zy.bms.redis.DeviceManager;
import com.zy.bms.service.*;
import com.zy.bms.utils.*;
import com.zy.bms.utils.CoordTransformUtil;
import com.zy.bms.utils.GaoDeApiUtil;
import com.zy.bms.websocket.WebSocketServer;
import org.springframework.http.HttpHeaders;
import org.springframework.stereotype.Component;

import javax.annotation.Resource;
import java.util.HashMap;
import java.util.Map;

/**
 * MQTT 消息返回处理类
 *
 * @author yang xiao kun
 * create on 2021/1/19
 */
@Component
public class MqttCallbackHandler {

    @Resource
    private DeviceService deviceService;
    @Resource
    private SetBaseService setBaseService;
    @Resource
    private SetToDeviceService setToDeviceService;
    @Resource
    private DeviceLocationService deviceLocationService;
    @Resource
    private DeviceManager deviceRedisService;
    @Resource
    private WebSocketServer webSocketServer;
    @Resource
    private CallRecordsService callRecordsService;
    @Resource
    private MqttLogService mqttLogService;

    /**
     * 处理消息
     *
     * @param topic   主题
     * @param payload 消息内容
     */
    void handle(String topic, String payload) {
        //存储日志
        mqttLogService.saveLog(topic, payload, 0);
        //消息返回JSON转Object
        MqttMsgDto msgObject = MqttMsgDto.parse(payload);
        //转发消息至网页
        webSocketServer.massMessage(msgObject.getDeviceId(), payload);
        //根据报文标识代码 M 处理消息
        switch (msgObject.getM()) {
            case Constant.M_CODE_REGISTER://设备端注册报文
                registerDeviceHandler(msgObject);
                break;
            case Constant.M_CODE_UPLOAD_LOCATION:// 设备位置信息上报
                uploadLocationHandler(msgObject);
                break;
            case Constant.M_CODE_UPLOAD_CALL_RECORD://设备上传通话记录
                uploadCallRecordsHandler(msgObject);
                break;
        }
    }

    /**
     * 注册设备
     */
    private void registerDeviceHandler(MqttMsgDto obj) {
        String num = obj.getString("regnum");
        //查询数据库中是否有此设备
        Device device = deviceService.getByNum(num);
        //没有该设备,保存至数据库
        if (device == null) {
            device = createDevice(num);
        }
        //mqtt服务器注册用户
        registerMqtt(device.getClientId(), device.getPassword());
        //发送消息
        setToDeviceService.returnCodeToDevice(device);
    }

    /**
     * 处理上传位置
     */
    private void uploadLocationHandler(MqttMsgDto msg) {
        DeviceLocation location = MqttMsgToLocation(msg);
        // 逆地理位置解析
        location.setSite(GaoDeApiUtil.regeo(location.getLonGcj(), location.getLatGcj()));
        // 保存位置信息--历史记录
        deviceLocationService.save(location);
        // 将最新位置信息保存至redis中
        deviceRedisService.saveLocation(location);
    }

    /**
     * 设备上传通话记录
     */
    private void uploadCallRecordsHandler(MqttMsgDto obj) {
        CallRecords callRecords = new CallRecords();
        callRecords.setDeviceId(obj.getDeviceId());
        callRecords.setCallType(obj.getInt("type"));
        callRecords.setKeyNum(obj.getString("key"));
        callRecords.setPhoneNum(obj.getString("phnoenum"));
        callRecords.setTalkTime(obj.getInt("talktime"));
        callRecords.setEndTime(obj.getString("endtime"));
        callRecordsService.save(callRecords);
    }

    /**
     * 创建设备信息,保存设备基础信息以及设置信息
     *
     * @param num 设备随机码
     */
    private Device createDevice(String num) {
        Device device = new Device();
        device.setNum(num);
        String clientId = MD5Util.MD5Encode(num);
        device.setClientId(clientId);
        device.setPassword(MD5Util.MD5Encode(clientId));
        device.setGroupId(1);
        device.setStatus(1);
        deviceService.save(device);
        SetBase setBase = new SetBase();
        setBase.setDeviceId(device.getClientId());
        setBaseService.save(setBase);
        return device;
    }

    /**
     * 解析信息转为位置对象
     */
    private DeviceLocation MqttMsgToLocation(MqttMsgDto obj) {
        try {
            DeviceLocation location = new DeviceLocation();
            location.setDeviceId(obj.getDeviceId());
            location.setBatteryNum(obj.getInt("batterynum"));
            location.setSignalNum(obj.getInt("signalnum"));
            location.setMode(obj.getInt("mode"));
            location.setLon(obj.getString("lon"));
            location.setLat(obj.getString("lat"));
            location.setSpeed(obj.getString("speed"));
            location.setNum(obj.getInt("num"));
            location.setUploadTime(obj.getDate("createtime"));
            Object cellInfo = obj.getObject("cellInfo");
            //该数据项可能为空
            if (cellInfo != null) location.setCellInfo(cellInfo.toString());
            //转换后的坐标
            String[] gcj = CoordTransformUtil.wgs84toGcj02(location.getLon(), location.getLat());
            location.setLonGcj(gcj[0]);
            location.setLatGcj(gcj[1]);
            return location;
        } catch (Exception e) {
            e.printStackTrace();
            throw new ApiRuntimeException("处理设备上传位置定位信息出错");
        }
    }


    /**
     * 向 MQTT 服务器申请用户,供设备进行登录
     * <p>
     * 因为存在先后问题,需要先注册用户,在开通虚拟机权限,
     * 所以逻辑为
     * 发送注册用户请求
     * 尝试申请权限
     * 如果失败,则间隔一秒再次请求
     * 如果申请次数超过3次,则尝试重新注册
     * 重新注册后再次尝试申请
     * 如果申请五次都失败,则结束方法,记录日志
     *
     * @param clientId 设备登录ID
     * @param password 设备登录密码
     */
    private void registerMqtt(String clientId, String password) {
        addRabbitMqUser(clientId, password);
        int count = 0;
        while (!setPermission(clientId)) {
            try {
                Thread.sleep(1000);
                if (count == 2) {
                    addRabbitMqUser(clientId, password);
                }
                if (count > 4) {
                    break;
                }
                count++;
            } catch (InterruptedException e) {
                e.printStackTrace();
                break;
            }
        }
    }

    /**
     * 注册 RabbitMQ 用户
     *
     * @param clientId 用户名
     * @param password 用户密码
     */
    private void addRabbitMqUser(String clientId, String password) {
        String url = "http://view.ailishi.org:15672/api/users/" + clientId;
        //请求头
        HttpHeaders headers = new HttpHeaders();
        headers.add("authorization", "Basic bHEyMDE5OkxpUXVhblJhYmJpdA==");
        //请求参数
        Map<String, String> params = new HashMap<>();
        params.put("username", clientId);
        params.put("password", password);
        params.put("tags", "");
        HttpUtil.putForJSONEntity(url, headers, JSON.toJSONString(params));
    }

    /**
     * 申请虚拟机权限
     *
     * @param clientId 用户ID
     */
    private boolean setPermission(String clientId) {
        String url = "http://view.ailishi.org:15672/api/permissions/%2F/" + clientId;
        //请求头
        HttpHeaders headers = new HttpHeaders();
        headers.add("authorization", "Basic bHEyMDE5OkxpUXVhblJhYmJpdA==");
        //请求参数
        Map<String, String> params = new HashMap<>();
        params.put("username", clientId);
        params.put("vhost", "/");
        params.put("configure", ".*");
        params.put("write", ".*");
        params.put("read", ".*");
        try {
            HttpUtil.putForJSONEntity(url, headers, JSON.toJSONString(params));
        } catch (Exception e) {
            return false;
        }
        return true;
    }
}
